from o3seespy.command.uniaxial_material.base_material import UniaxialMaterialBase


class PySimple2(UniaxialMaterialBase):
    """
    The PySimple2 UniaxialMaterial Class

    This command is used to construct a PySimple2 uniaxial material object.
    """
    op_type = 'PySimple2'

    def __init__(self, osi, soil_type, pult, y50, cd, c=0.0):
        """
        Initial method for PySimple2

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        soil_type: int
            Soiltype = 1 backbone of p-y curve approximates matlock (1970) soft clay relation. soiltype = 2 backbone of
            p-y curve approximates api (1993) sand relation.
        pult: float
            Ultimate capacity of the p-y material. note that "p" or "pult" are distributed loads [force per length of
            pile] in common design equations, but are both loads for this uniaxial_material [i.e., distributed load times the
            tributary length of the pile].
        y50: float
            Displacement at which 50% of pult is mobilized in monotonic loading.
        cd: float
            Variable that sets the drag resistance within a fully-mobilized gap as cd*pult.
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            (optional default = 0.0). nonzero c values are used to represent radiation damping effects

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.PySimple2(osi, soil_type=1, pult=1.0, y50=1.0, cd=1.0, c=0.0)
        """
        self.osi = osi
        self.soil_type = int(soil_type)
        self.pult = float(pult)
        self.y50 = float(y50)
        self.cd = float(cd)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.soil_type, self.pult, self.y50, self.cd, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class QzSimple2(UniaxialMaterialBase):
    """
    The QzSimple2 UniaxialMaterial Class

    This command is used to construct a QzSimple2 uniaxial material object.
    """
    op_type = 'QzSimple2'

    def __init__(self, osi, qz_type, qult, z50, suction=0.0, c=0.0):
        """
        Initial method for QzSimple2

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        qz_type: int
            Qztype = 1 backbone of q-z curve approximates reese and o'neill's (1987) relation for drilled shafts in
            clay. qztype = 2 backbone of q-z curve approximates vijayvergiya's (1977) relation for piles in sand.
        qult: float
            Ultimate capacity of the q-z material. see note 1.
        z50: float
            Displacement at which 50% of qult is mobilized in monotonic loading. see note 2.
        suction: float, optional
            Uplift resistance is equal to suction*qult. default = 0.0. the value of suction must be 0.0 to 0.1.*
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            default = 0.0. nonzero c values are used to represent radiation damping effects.*

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.QzSimple2(osi, qz_type=1, qult=1.0, z50=1.0, suction=0.0, c=0.0)
        """
        self.osi = osi
        self.qz_type = int(qz_type)
        self.qult = float(qult)
        self.z50 = float(z50)
        self.suction = float(suction)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.qz_type, self.qult, self.z50, self.suction, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TzSimple2(UniaxialMaterialBase):
    """
    The TzSimple2 UniaxialMaterial Class

    This command is used to construct a TzSimple2 uniaxial material object.
    """
    op_type = 'TzSimple2'

    def __init__(self, osi, soil_type, tult, z50, c=0.0):
        """
        Initial method for TzSimple2

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        soil_type: int
            Soiltype = 1 backbone of t-z curve approximates reese and o'neill (1987). soiltype = 2 backbone of t-z curve
            approximates mosher (1984) relation.
        tult: float
            Ultimate capacity of the t-z material. see note 1.
        z50: float
            Displacement at which 50% of tult is mobilized in monotonic loading.
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            (optional default = 0.0). see note 2.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.TzSimple2(osi, soil_type=1, tult=1.0, z50=1.0, c=0.0)
        """
        self.osi = osi
        self.soil_type = int(soil_type)
        self.tult = float(tult)
        self.z50 = float(z50)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.soil_type, self.tult, self.z50, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



from ..node import Node
from ..element.zero_length import ZeroLength
from .beam_column import ElasticBeamColumn2D
from ..geom_transf import Linear2D as geom_trans_Linear2D
from .bearing import FlatSliderBearing2D
from ..uniaxial_material import Elastic
from ... import cc
from .. import common


class BeamOnNonlinearWinklerFoundation(object):
    def __init__(self, top_nodes, bot_nodes, sf_eles, fd_eles, sf_mats=None, sf_horz_eles=None, sf_horz_mats=None):
        self.top_nodes = top_nodes
        self.bot_nodes = bot_nodes
        self.sf_eles = sf_eles
        self.sf_horz_eles = sf_horz_eles
        self.fd_eles = fd_eles
        self.sf_mats = sf_mats
        self.sf_horz_mats = sf_horz_mats


def gen_shallow_foundation_bnwf(osi, bottom_node, top_node, sf_mats, pos, fd_area, fd_e_mod, fd_iz,
                                    r_flag: float = None, orient: list = None, sf_horz_mats=None, sf_frn=None, offset=0.0):
    """
    Generates nodes and elements for a shallow foundation beam on nonlinear foundation

    Alpha Status - Subject to changes and testing

    Note: Only supports 2D

    Parameters
    ----------
    osi
    bottom_node
    top_node
    sf_mats
    pos
    fd_area
    fd_e_mod
    fd_iz
    r_flag
    orient

    Returns
    -------

    """
    # TODO: sf_horz_mats - if 1 then apply to left node, if len(sf_mats) apply to each
    # TODO: sf_frn - soil-foundation friction, if float, then add each sf_mat inside a flat bearing.
    if not hasattr(sf_mats, '__len__'):  # TODO: only supports 2D
        sf_mats = [sf_mats] * len(pos)
    else:
        sf_mats = list(sf_mats)
    if hasattr(sf_horz_mats, '__len__'):
        assert len(sf_mats) == len(sf_horz_mats)
        sf_horz1_mat = None
    else:
        sf_horz1_mat = sf_horz_mats
        sf_horz_mats = [None] * len(sf_mats)
    injected = 0
    pos = list(pos)
    if offset not in pos:
        pos.append(offset)
        pos.sort()
        injected = 1
    ind = pos.index(offset)
    if injected:
        sf_mats.insert(ind, None)  # insert a null material since there should not be a soil spring here
        sf_horz_mats.insert(ind, None)
    top_nodes = []
    bot_nodes = []
    frn_nodes = []  # unused if no friction model
    con_nodes = top_nodes  # symlink to bot_nodes
    y_inc = 0
    if sf_frn:
        con_nodes = frn_nodes
        y_inc = 0.001  # need small vertical offset for friction element
    sf_eles = []
    sf_horz_eles = []
    fd_eles = []
    transf = geom_trans_Linear2D(osi, [])
    x_centre = bottom_node.x
    y_centre = bottom_node.y
    for i, x in enumerate(pos):
        top_nodes.append(Node(osi, x_centre + x, y_centre))  # node that connects to foundation element
        bot_nodes.append(Node(osi, x_centre + x, y_centre - y_inc))  # node that connects to the far field soil (fixed)
        common.Fix3DOF(osi, bot_nodes[i], cc.FIXED, cc.FIXED, cc.FIXED)
        if sf_frn:  # intermediate nodes to connect friction model between top_nodes and frn_nodes
            frn_nodes.append(Node(osi, x_centre + x, y_centre - y_inc))
            k_add_shear = 1e12
            p_mat = Elastic(osi, 1e12)
            ele = FlatSliderBearing2D(osi, [top_nodes[i], frn_nodes[i]], sf_frn, k_add_shear, p_mat=p_mat,
                                                 mz_mat=p_mat)  # Pretty sure the p_mat can be None?
        if sf_mats[i] is not None:
            sf_eles.append(ZeroLength(osi, [con_nodes[i], bot_nodes[i]], [sf_mats[i]], [cc.DOF2D_Y], r_flag, orient))
        if sf_horz_mats[i] is not None:
            sf_horz_eles.append(ZeroLength(osi, [con_nodes[i], bot_nodes[i]], [sf_horz_mats[i]], [cc.DOF2D_X], r_flag, orient))
        if i != 0:  # define foundation elements between the top_nodes
            fd_eles.append(ElasticBeamColumn2D(osi, [top_nodes[i - 1], top_nodes[i]], fd_area, fd_e_mod, fd_iz, transf))
    common.EqualDOF(osi, top_node, top_nodes[ind], dofs=[cc.DOF2D_X, cc.DOF2D_Y, cc.DOF2D_ROTZ])
    if sf_horz1_mat:
        sf_horz_mats = ZeroLength(osi, [top_nodes[0], bot_nodes[0]], [sf_horz1_mat], [cc.DOF2D_X], r_flag, orient)
    return BeamOnNonlinearWinklerFoundation(top_nodes, bot_nodes, sf_eles, fd_eles, sf_mats, sf_horz_eles, sf_horz_mats)



from o3seespy.command.element.base_element import ElementBase


class Quad(ElementBase):
    """
    The Quad Element Class
    
    This command is used to construct a FourNodeQuad element object which uses a bilinear isoparametric formulation.

  
     
    """
    op_type = 'quad'

    def __init__(self, osi, ele_nodes, thick, otype, mat, pressure=0.0, rho=0.0, b1=0.0, b2=0.0):
        """
        Initial method for Quad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        otype: str
            String representing material behavior. the type parameter can be either ``'planestrain'`` or
            ``'planestress'``
        mat: obj
            Object of ndmaterial
        pressure: float, optional
            Surface pressure (optional, default = 0.0)
        rho: float, optional
            Element mass density (per unit volume) from which a lumped element mass matrix is computed (optional,
            default=0.0)
        b1: float, optional
            Constant body forces defined in the isoparametric domain (optional, default=0.0)
        b2: float, optional
            Constant body forces defined in the isoparametric domain (optional, default=0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.Quad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStrain', mat=obj, pressure=1.0, rho=1.0, b1=0.0, b2=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.otype = otype
        self.mat = mat
        self.pressure = float(pressure)
        self.rho = float(rho)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.otype, self.mat.tag, self.pressure, self.rho, self.b1, self.b2]
        self.to_process(osi)


class ShellMITC4(ElementBase):
    """
    The ShellMITC4 Element Class
    
    This command is used to construct a ShellMITC4 element object, which uses a bilinear isoparametric formulation in
    combination with a modified shear interpolation to improve thin-plate bending performance.

    
    """
    op_type = 'ShellMITC4'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellMITC4

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently must be either a
            ``'platefibersection'``, or ``'elasticmembraneplatesection'``

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellMITC4(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellDKGQ(ElementBase):
    """
    The ShellDKGQ Element Class
    
    This command is used to construct a ShellDKGQ element object, which is a quadrilateral shell element based on the
    theory of generalized conforming element.

    
    """
    op_type = 'ShellDKGQ'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellDKGQ

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellDKGQ(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellDKGT(ElementBase):
    """
    The ShellDKGT Element Class
    
    This command is used to construct a ShellDKGT element object, which is a triangular shell element based on the
    theory of generalized conforming element.

    
    """
    op_type = 'ShellDKGT'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellDKGT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of three element nodes in clockwise or counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellDKGT(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNLDKGQ(ElementBase):
    """
    The ShellNLDKGQ Element Class
    
    This command is used to construct a ShellNLDKGQ element object accounting for the geometric nonlinearity of large
    deformation using the updated Lagrangian formula, which is developed based on the ShellDKGQ element.

    
    """
    op_type = 'ShellNLDKGQ'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNLDKGQ

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNLDKGQ(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNLDKGT(ElementBase):
    """
    The ShellNLDKGT Element Class
    
    This command is used to construct a ShellNLDKGT element object accounting for the geometric nonlinearity of large
    deformation using the updated Lagrangian formula, which is developed based on the ShellDKGT element.

    
    """
    op_type = 'ShellNLDKGT'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNLDKGT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of three element nodes in clockwise or counter-clockwise order around the element
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNLDKGT(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNL(ElementBase):
    """
    The ShellNL Element Class
    
    
    """
    op_type = 'ShellNL'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNL

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of nine element nodes, input is the typical, firstly four corner nodes counter-clockwise, then
            mid-side nodes counter-clockwise and finally the central node
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1], [0.5, 0], [1, 0.5], [0.5, 1], [0, 0.5], [0.5, 0.5]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNL(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class BbarQuad(ElementBase):
    """
    The BbarQuad Element Class
    
    This command is used to construct a four-node quadrilateral element object, which uses a bilinear isoparametric
    formulation along with a mixed volume/pressure B-bar assumption. This element is for plane strain problems only.

    
    """
    op_type = 'bbarQuad'

    def __init__(self, osi, ele_nodes, thick, mat):
        """
        Initial method for BbarQuad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        mat: obj
            Object of ndmaterial

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.BbarQuad(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.mat.tag]
        self.to_process(osi)


class EnhancedQuad(ElementBase):
    """
    The EnhancedQuad Element Class
    
    This command is used to construct a four-node quadrilateral element, which uses a bilinear isoparametric formulation
    with enhanced strain modes.

    
    """
    op_type = 'enhancedQuad'

    def __init__(self, osi, ele_nodes, thick, otype, mat):
        """
        Initial method for EnhancedQuad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        otype: str
            String representing material behavior. valid options depend on the ndmaterial object and its available
            material formulations. the type parameter can be either ``'planestrain'`` or ``'planestress'``
        mat: obj
            Object of ndmaterial

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.EnhancedQuad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStress', mat=obj)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.otype = otype
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.otype, self.mat.tag]
        self.to_process(osi)


class SSPquad(ElementBase):
    """
    The SSPquad Element Class
    
    This command is used to construct a SSPquad element object.

    
    """
    op_type = 'SSPquad'

    def __init__(self, osi, ele_nodes, mat, otype, thick, b1, b2):
        """
        Initial method for SSPquad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        otype: str
            String to relay material behavior to the element, can be either ``'planestrain'`` or ``'planestress'``
        thick: float
            Thickness of the element in out-of-plane direction
        b1: float
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0)
        b2: float
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.SSPquad(osi, ele_nodes=ele_nodes, mat=obj, otype='PlaneStrain', thick=1.0, b1=0.0, b2=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.otype = otype
        self.thick = float(thick)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.otype, self.thick, self.b1, self.b2]
        self.to_process(osi)


class MVLEM3DCoR(ElementBase):
    """
    The MVLEM3DCoR Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The MVLEM_3D model (Figure 1a) is a
    three-dimensional four-node element with 24 DOFs for nonlinear analysis of flexure-controlled non-rectangular
    reinforced concrete walls subjected to multidirectional loading. The model is an extension of the
    two-dimensional, two-node Multiple-Vertical-Line-Element-Model (`MVLEM
    
    """
    op_type = 'MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, c, thick: list=None, widths: list=None, rho: list=None, mat_concretes: list=None, mat_steels: list=None, mat_shear=None):
        r"""
        Initial method for MVLEM3DCoR

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        c: float
            Location of center of rotation from the base (optional; default = 0.4 (recommended))
        thick: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        rho: list, optional
            A list of m reinforcing ratios corresponding to macro-fibers; for each fiber: :math:`rho_i =
            a_{s,i}/a_{gross,i} (1 < i < m)`
        mat_concretes: list, optional
            A list of ``m`` uniaxial_material objects for concrete
        mat_steels: list, optional
            A list of ``m`` uniaxial_material objects for steel
        mat_shear: obj, optional
            Object of uniaxial_material for shear material
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thick = thick
        self.widths = widths
        self.rho = rho
        if mat_concretes is None:
            self.mat_concretes = None
        else:
            self.mat_concretes = [x.tag for x in mat_concretes]
        if mat_steels is None:
            self.mat_steels = None
        else:
            self.mat_steels = [x.tag for x in mat_steels]
        self.mat_shear = mat_shear
        self.c = float(c)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-CoR', self.c]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'rho') is not None:
            self._parameters += ['-rho', *self.rho]
        if getattr(self, 'mat_concretes') is not None:
            self._parameters += ['-matConcrete', *self.mat_concretes]
        if getattr(self, 'mat_steels') is not None:
            self._parameters += ['-matSteel', *self.mat_steels]
        if getattr(self, 'mat_shear') is not None:
            self._parameters += ['-matShear', self.mat_shear.tag]
        self.to_process(osi)

class MVLEM3DThickMod(ElementBase):
    """
    The MVLEM3DThickMod Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The MVLEM_3D model (Figure 1a) is a
    three-dimensional four-node element with 24 DOFs for nonlinear analysis of flexure-controlled non-rectangular
    reinforced concrete walls subjected to multidirectional loading. The model is an extension of the
    two-dimensional, two-node Multiple-Vertical-Line-Element-Model (`MVLEM
    
    """
    op_type = 'MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, t_mod, thick: list=None, widths: list=None, rho: list=None, mat_concretes: list=None, mat_steels: list=None, mat_shear=None):
        r"""
        Initial method for MVLEM3DThickMod

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        t_mod: float
            Thickness multiplier (optional; default = 0.63 equivalent to 0.25ig for out-of-plane bending)
        thick: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        rho: list, optional
            A list of m reinforcing ratios corresponding to macro-fibers; for each fiber: :math:`rho_i =
            a_{s,i}/a_{gross,i} (1 < i < m)`
        mat_concretes: list, optional
            A list of ``m`` uniaxial_material objects for concrete
        mat_steels: list, optional
            A list of ``m`` uniaxial_material objects for steel
        mat_shear: obj, optional
            Object of uniaxial_material for shear material
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thick = thick
        self.widths = widths
        self.rho = rho
        if mat_concretes is None:
            self.mat_concretes = None
        else:
            self.mat_concretes = [x.tag for x in mat_concretes]
        if mat_steels is None:
            self.mat_steels = None
        else:
            self.mat_steels = [x.tag for x in mat_steels]
        self.mat_shear = mat_shear
        self.t_mod = float(t_mod)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-ThickMod', self.t_mod]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'rho') is not None:
            self._parameters += ['-rho', *self.rho]
        if getattr(self, 'mat_concretes') is not None:
            self._parameters += ['-matConcrete', *self.mat_concretes]
        if getattr(self, 'mat_steels') is not None:
            self._parameters += ['-matSteel', *self.mat_steels]
        if getattr(self, 'mat_shear') is not None:
            self._parameters += ['-matShear', self.mat_shear.tag]
        self.to_process(osi)

class MVLEM3DPoisson(ElementBase):
    """
    The MVLEM3DPoisson Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The MVLEM_3D model (Figure 1a) is a
    three-dimensional four-node element with 24 DOFs for nonlinear analysis of flexure-controlled non-rectangular
    reinforced concrete walls subjected to multidirectional loading. The model is an extension of the
    two-dimensional, two-node Multiple-Vertical-Line-Element-Model (`MVLEM
    
    """
    op_type = 'MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, nu, thick: list=None, widths: list=None, rho: list=None, mat_concretes: list=None, mat_steels: list=None, mat_shear=None):
        r"""
        Initial method for MVLEM3DPoisson

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        nu: float
            Poisson ratio for out-of-plane bending (optional; default = 0.25)
        thick: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        rho: list, optional
            A list of m reinforcing ratios corresponding to macro-fibers; for each fiber: :math:`rho_i =
            a_{s,i}/a_{gross,i} (1 < i < m)`
        mat_concretes: list, optional
            A list of ``m`` uniaxial_material objects for concrete
        mat_steels: list, optional
            A list of ``m`` uniaxial_material objects for steel
        mat_shear: obj, optional
            Object of uniaxial_material for shear material
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thick = thick
        self.widths = widths
        self.rho = rho
        if mat_concretes is None:
            self.mat_concretes = None
        else:
            self.mat_concretes = [x.tag for x in mat_concretes]
        if mat_steels is None:
            self.mat_steels = None
        else:
            self.mat_steels = [x.tag for x in mat_steels]
        self.mat_shear = mat_shear
        self.nu = float(nu)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-Poisson', self.nu]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'rho') is not None:
            self._parameters += ['-rho', *self.rho]
        if getattr(self, 'mat_concretes') is not None:
            self._parameters += ['-matConcrete', *self.mat_concretes]
        if getattr(self, 'mat_steels') is not None:
            self._parameters += ['-matSteel', *self.mat_steels]
        if getattr(self, 'mat_shear') is not None:
            self._parameters += ['-matShear', self.mat_shear.tag]
        self.to_process(osi)

class MVLEM3DDensity(ElementBase):
    """
    The MVLEM3DDensity Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The MVLEM_3D model (Figure 1a) is a
    three-dimensional four-node element with 24 DOFs for nonlinear analysis of flexure-controlled non-rectangular
    reinforced concrete walls subjected to multidirectional loading. The model is an extension of the
    two-dimensional, two-node Multiple-Vertical-Line-Element-Model (`MVLEM
    
    """
    op_type = 'MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, dens, thick: list=None, widths: list=None, rho: list=None, mat_concretes: list=None, mat_steels: list=None, mat_shear=None):
        r"""
        Initial method for MVLEM3DDensity

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        dens: float
            Density (optional; default = 0.0)
        thick: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        rho: list, optional
            A list of m reinforcing ratios corresponding to macro-fibers; for each fiber: :math:`rho_i =
            a_{s,i}/a_{gross,i} (1 < i < m)`
        mat_concretes: list, optional
            A list of ``m`` uniaxial_material objects for concrete
        mat_steels: list, optional
            A list of ``m`` uniaxial_material objects for steel
        mat_shear: obj, optional
            Object of uniaxial_material for shear material
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thick = thick
        self.widths = widths
        self.rho = rho
        if mat_concretes is None:
            self.mat_concretes = None
        else:
            self.mat_concretes = [x.tag for x in mat_concretes]
        if mat_steels is None:
            self.mat_steels = None
        else:
            self.mat_steels = [x.tag for x in mat_steels]
        self.mat_shear = mat_shear
        self.dens = float(dens)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-Density', self.dens]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'rho') is not None:
            self._parameters += ['-rho', *self.rho]
        if getattr(self, 'mat_concretes') is not None:
            self._parameters += ['-matConcrete', *self.mat_concretes]
        if getattr(self, 'mat_steels') is not None:
            self._parameters += ['-matSteel', *self.mat_steels]
        if getattr(self, 'mat_shear') is not None:
            self._parameters += ['-matShear', self.mat_shear.tag]
        self.to_process(osi)


class SFIMVLEM3DCoR(ElementBase):
    """
    The SFIMVLEM3DCoR Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The SFI-MVLEM-3D model (Figure 1a)
    is a three-dimensional four-node element with 24 DOFs that incorporates axial-flexural-shear interaction and can
    be used for nonlinear analysis of non-rectangular reinforced concrete walls subjected to multidirectional
    loading. The SFI-MVLEM-3D model is an extension of the two-dimensional, two-node
    Shear-Flexure-Interaction Multiple-Vertical-Line-Element-Model (`SFI-MVLEM
    
    """
    op_type = 'SFI_MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, c, thicks: list=None, widths: list=None, mats: list=None):
        """
        Initial method for SFIMVLEM3DCoR

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        c: float
            Location of center of rotation from the base (optional; default = 0.4 (recommended))
        thicks: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        mats: list, optional
            A list of ``m`` material objects corresponding to ndmaterial fsam
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thicks = thicks
        self.widths = widths
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.c = float(c)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-CoR', self.c]
        if getattr(self, 'thicks') is not None:
            self._parameters += ['-thick', *self.thicks]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        self.to_process(osi)

class SFIMVLEM3DThickMod(ElementBase):
    """
    The SFIMVLEM3DThickMod Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The SFI-MVLEM-3D model (Figure 1a)
    is a three-dimensional four-node element with 24 DOFs that incorporates axial-flexural-shear interaction and can
    be used for nonlinear analysis of non-rectangular reinforced concrete walls subjected to multidirectional
    loading. The SFI-MVLEM-3D model is an extension of the two-dimensional, two-node
    Shear-Flexure-Interaction Multiple-Vertical-Line-Element-Model (`SFI-MVLEM
    
    """
    op_type = 'SFI_MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, t_mod, thicks: list=None, widths: list=None, mats: list=None):
        """
        Initial method for SFIMVLEM3DThickMod

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        t_mod: float
            Thickness multiplier (optional; default = 0.63 equivalent to 0.25ig for out-of-plane bending)
        thicks: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        mats: list, optional
            A list of ``m`` material objects corresponding to ndmaterial fsam
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thicks = thicks
        self.widths = widths
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.t_mod = float(t_mod)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-ThickMod', self.t_mod]
        if getattr(self, 'thicks') is not None:
            self._parameters += ['-thick', *self.thicks]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        self.to_process(osi)

class SFIMVLEM3DPoisson(ElementBase):
    """
    The SFIMVLEM3DPoisson Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The SFI-MVLEM-3D model (Figure 1a)
    is a three-dimensional four-node element with 24 DOFs that incorporates axial-flexural-shear interaction and can
    be used for nonlinear analysis of non-rectangular reinforced concrete walls subjected to multidirectional
    loading. The SFI-MVLEM-3D model is an extension of the two-dimensional, two-node
    Shear-Flexure-Interaction Multiple-Vertical-Line-Element-Model (`SFI-MVLEM
    
    """
    op_type = 'SFI_MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, nu, thicks: list=None, widths: list=None, mats: list=None):
        """
        Initial method for SFIMVLEM3DPoisson

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        nu: float
            Poisson ratio for out-of-plane bending (optional; default = 0.25)
        thicks: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        mats: list, optional
            A list of ``m`` material objects corresponding to ndmaterial fsam
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thicks = thicks
        self.widths = widths
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.nu = float(nu)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-Poisson', self.nu]
        if getattr(self, 'thicks') is not None:
            self._parameters += ['-thick', *self.thicks]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        self.to_process(osi)

class SFIMVLEM3DDensity(ElementBase):
    """
    The SFIMVLEM3DDensity Element Class
    
    | Developed and implemented by: | `Kristijan Kolozvari <mailto:kkolozvari@fullerton.edu>`_ (CSU Fullerton)| Kamiar
    Kalbasi (CSU Fullerton)| Kutay Orakcal (Bogazici University)| John Wallace (UCLA)The SFI-MVLEM-3D model (Figure 1a)
    is a three-dimensional four-node element with 24 DOFs that incorporates axial-flexural-shear interaction and can
    be used for nonlinear analysis of non-rectangular reinforced concrete walls subjected to multidirectional
    loading. The SFI-MVLEM-3D model is an extension of the two-dimensional, two-node
    Shear-Flexure-Interaction Multiple-Vertical-Line-Element-Model (`SFI-MVLEM
    
    """
    op_type = 'SFI_MVLEM_3D'

    def __init__(self, osi, ele_nodes, m, dens, thicks: list=None, widths: list=None, mats: list=None):
        """
        Initial method for SFIMVLEM3DDensity

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes defined in the counter-clockwise direction
        m: int
            Number of element uniaxial fibers
        dens: float
            Density (optional; default = 0.0)
        thicks: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        mats: list, optional
            A list of ``m`` material objects corresponding to ndmaterial fsam
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.thicks = thicks
        self.widths = widths
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.dens = float(dens)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, '-Density', self.dens]
        if getattr(self, 'thicks') is not None:
            self._parameters += ['-thick', *self.thicks]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class AlgorithmBase(OpenSeesObject):
    op_base_type = "algorithm"

    def to_process(self, osi):
        if osi is None:
            return
        OpenSeesObject.to_process(self, osi)

    def reapply(self, osi):
        self.to_process(osi)


class Linear(AlgorithmBase):
    op_type = "Linear"

    def __init__(self, osi, secant=False, initial=False, factor_once=False):
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self.factor_once = factor_once
        self._parameters = [self.op_type, self.secant, self.initial, self.factor_once]
        self.to_process(osi)


class Newton(AlgorithmBase):
    op_type = "Newton"

    def __init__(self, osi, secant=False, initial=False, initial_then_current=False):
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self.initial_then_current = initial_then_current
        self._parameters = [self.op_type, self.secant, self.initial, self.initial_then_current]
        self.to_process(osi)


class SecantNewton(AlgorithmBase):
    op_type = "SecantNewton"

    def __init__(self, osi, iterate='current', increment='current', maxDim=3):
        self.osi = osi
        self.iterate = iterate
        self.increment = increment
        self.maxDim = maxDim
        self._parameters = [self.op_type, self.iterate, self.increment, self.maxDim]
        self.to_process(osi)


class RaphsonNewton(AlgorithmBase):
    op_type = "RaphsonNewton"

    def __init__(self, osi, iterate='current', increment='current'):
        self.osi = osi
        self.iterate = iterate
        self.increment = increment
        self._parameters = [self.op_type, self.iterate, self.increment]
        self.to_process(osi)


class KrylovNewton(AlgorithmBase):
    op_type = "KrylovNewton"

    def __init__(self, osi, tang_inter='current', tang_incr='current', max_inter=3):
        """

        Parameters
        ----------
        osi
        tang_inter: str
            options are: 'current', 'initial', 'noTangent'
        tang_incr
        max_inter
        """
        self.osi = osi
        self.tang_inter = tang_inter
        self.tang_incr = tang_incr
        self.max_inter = max_inter
        self._parameters = [self.op_type, self.tang_inter, self.tang_incr, self.max_inter]
        self.to_process(osi)


class ModifiedNewton(AlgorithmBase):
    op_type = "ModifiedNewton"

    def __init__(self, osi, secant=False, initial=False):
        """

        Parameters
        ----------
        osi
        secant: bool
            Flag to indicate to use secant stiffness.
        initial: bool
            Flag to indicate to use initial stiffness
        max_inter
        """
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self._parameters = [self.op_type, self.secant, self.initial]
        self.to_process(osi)


class NewtonLineSearch(AlgorithmBase):
    op_type = 'NewtonLineSearch'

    def __init__(self, osi, search_type=None, tol=None, max_iter=None, min_eta=None, max_eta=None):
        """
        
        :param osi: 
        :param search_type: str 
        :param tol: default 0.8
        :param max_iter: default=10
        :param min_eta: deault=0.1
        :param max_eta: default=10
        """
        self.osi = osi
        if search_type is None:
            self.search_type = None
            self._parameters = [self.op_type, search_type]
        else:
            self.search_type = search_type
            self._parameters = [self.op_type, search_type]
        
        if tol is not None:
            self._parameters += ['-tol', tol]
        if max_iter is not None:
            self._parameters += ['-maxIter', max_iter]
        if min_eta is not None:
            self._parameters += ['-minEta', min_eta]
        if max_eta is not None:
            self._parameters += ['-maxEta', max_eta]
        self.to_process(osi)


class Broyden(AlgorithmBase):
    op_type = 'Broyden'

    def __init__(self, osi, secant=False, initial=False, count=10):
        self.osi = osi
        self._parameters = [self.op_type, secant, initial, count]
        self.to_process(osi)

class BFGS(AlgorithmBase):
    op_type = 'BFGS'

    def __init__(self, osi, secant=False, initial=False, count=10):
        self.osi = osi
        self._parameters = [self.op_type, secant, initial, count]
        self.to_process(osi)


import os
import o3seespy as o3


def calc_quad_centroids(xs, ys, axis=-1):  # use axis=-2 for time domain data, since -1 is time
    import numpy as np
    x0 = np.array(xs)
    y0 = np.array(ys)
    x1 = np.roll(xs, 1, axis=axis)
    y1 = np.roll(ys, 1, axis=axis)
    a = x0 * y1 - x1 * y0
    xc = np.sum((x0 + x1) * a, axis=axis)
    yc = np.sum((y0 + y1) * a, axis=axis)

    area = 0.5 * np.sum(a, axis=axis)
    xc /= (6.0 * area)
    yc /= (6.0 * area)

    return xc, yc


class Results2D(object):
    coords = None
    time = None
    x_disp = None
    y_disp = None
    node_c = None
    ele_c = None
    selected_nodes = None
    used_r_starter = 0
    mat2ele_tags = None  # Assume 1-to-1 so if it uses a section then should be null
    sect2ele_tags = None  # Store position and tag - UNUSED
    mat2sect_tags = None  # UNUSED  # TODO: implement
    n_nodes_per_ele = [2, 4, 8]  # for 2D
    ele_num_base = 1
    selected_nodes = None
    _selected_node_tags = None

    def __init__(self, cache_path='', dt=None, dynamic=False, man_nodes=False):
        self.cache_path = cache_path
        self._dt = dt
        self.dynamic = dynamic
        from numpy import savetxt, loadtxt
        self.savetxt = savetxt
        self.loadtxt = loadtxt
        self.ele2node_tags = {}
        self.meta_files = ['node_c', 'ele_c', 'mat2ele_tags', 'sect2ele_tags', 'mat2sect_tags']
        self.meta_fmt = [None, '%i', '%i', '%i']
        self.pseudo_dt = None  # use if recording steps of a static analysis
        self.man_nodes = man_nodes

    def start_recorders(self, osi, dt=None):  # TODO: handle recorder time step
        self.used_r_starter = 1
        if self.man_nodes and self.selected_node_tags is None:
            self._selected_node_tags = o3.get_node_tags(osi)
        if self.coords is None:
            if self.selected_node_tags is not None:
                self.coords = []
                for node_tag in self.selected_node_tags:
                    self.coords.append(o3.get_node_coords(osi, node_tag, node_as_tag=True))
            else:
                self.coords = o3.get_all_node_coords(osi)
        if not self.ele2node_tags:
            self.ele2node_tags = o3.get_all_ele_node_tags_as_dict(osi)
        if dt is not None:
            self._dt = dt
        if self.dynamic:
            node_tags = 'all'
            if self.selected_node_tags is not None:
                node_tags = self.selected_node_tags
            o3.recorder.NodesToFile(osi, f'{self.cache_path}x_disp.txt', node_tags, [o3.cc.DOF2D_X], 'disp', nsd=4, dt=dt, nodes_as_tags=True)
            o3.recorder.NodesToFile(osi, f'{self.cache_path}y_disp.txt', node_tags, [o3.cc.DOF2D_Y], 'disp', nsd=4, dt=dt, nodes_as_tags=True)
            if not self.pseudo_dt:
                o3.recorder.TimeToFile(osi, f'{self.cache_path}timer.txt', nsd=4, dt=dt)

    def wipe_old_files(self):
        sfiles = ['ele2node_tags', 'coords', 'selected_node_tags']
        if not self.used_r_starter:
            sfiles += ['x_disp', 'y_disp', 'timer']
        for fname in sfiles:
            try:
                os.remove(f'{self.cache_path}{fname}.txt')
            except FileNotFoundError:
                pass

        for fname in self.meta_files:
            try:
                os.remove(f'{self.cache_path}{fname}.txt')
            except FileNotFoundError:
                pass

    @property
    def selected_node_tags(self):
        if self._selected_node_tags is None:
            if self.selected_nodes is None:
                return None
            self._selected_node_tags = [x.tag for x in self.selected_nodes]
        return self._selected_node_tags

    @selected_node_tags.setter
    def selected_node_tags(self, tags):
        self._selected_node_tags = tags

    def save_to_cache(self):
        self.wipe_old_files()
        if self.coords is not None:
            self.savetxt(self.cache_path + 'coords.txt', self.coords)
        ostr = [f'{ele_tag} ' + ' '.join([str(x) for x in self.ele2node_tags[ele_tag]]) + '\n' for ele_tag in self.ele2node_tags]
        open(self.cache_path + 'ele2node_tags.txt', 'w').writelines(ostr)
        if self.selected_node_tags is not None:
            self.savetxt(self.cache_path + 'selected_node_tags.txt', self.selected_node_tags, fmt='%i')

        for i, fname in enumerate(self.meta_files):
            vals = getattr(self, fname)
            if vals is not None:
                self.savetxt(self.cache_path + f'{fname}.txt', vals, fmt=self.meta_fmt[i])
        if self.dynamic:
            if not self.used_r_starter:
                self.savetxt(self.cache_path + 'x_disp.txt', self.x_disp)
                self.savetxt(self.cache_path + 'y_disp.txt', self.y_disp)
                self.savetxt(self.cache_path + 'timer.txt', self.time)
            elif self.pseudo_dt:
                from numpy import arange
                x_disp = self.loadtxt(f'{self.cache_path}x_disp.txt', ndmin=2)
                self.time = arange(len(x_disp[:, 0])) * self.pseudo_dt
                self.savetxt(self.cache_path + 'timer.txt', self.time)

    def load_from_cache(self):
        self.coords = self.loadtxt(self.cache_path + 'coords.txt')

        try:
            self.selected_node_tags = self.loadtxt(self.cache_path + 'selected_node_tags.txt', dtype=int)
        except OSError:
            pass

        self.ele2node_tags = {}
        lines = open(self.cache_path + 'ele2node_tags.txt').read().splitlines()
        for line in lines:
            parts = [int(x) for x in line.split()]
            self.ele2node_tags[parts[0]] = parts[1:]
        for fname in self.meta_files:
            try:
                data = self.loadtxt(self.cache_path + f'{fname}.txt')
                if len(data) == 0:
                    data = None
                setattr(self, fname, data)
            except OSError:
                pass

        if self.dynamic:
            self.x_disp = self.loadtxt(f'{self.cache_path}x_disp.txt')
            self.y_disp = self.loadtxt(f'{self.cache_path}y_disp.txt')
            self.time = self.loadtxt(f'{self.cache_path}timer.txt', ndmin=2)[:, 0]

    @property
    def dt(self):
        if self._dt is None:
            if self.time is not None:
                self._dt = (self.time[-1] - self.time[0]) / (len(self.time) - 1)
        return self._dt

    @dt.setter
    def dt(self, dt):
        self._dt = dt

    def rezero_node_tags(self, osi=None):
        from numpy import array, arange, searchsorted, where
        if self.selected_node_tags is None:
            node_tags = array(o3.get_node_tags(osi))
        else:
            node_tags = self.selected_node_tags
        new_node_tags = arange(1, len(node_tags) + 1)
        sidx = node_tags.argsort()
        k = node_tags[sidx]
        v = new_node_tags[sidx]
        for ele_tag in self.ele2node_tags:
            curr_tags = self.ele2node_tags[ele_tag]
            idx = searchsorted(k, curr_tags)
            assert max(idx) < len(k)
            mask = k[idx] == curr_tags
            self.ele2node_tags[ele_tag] = where(mask, v[idx], len(k))


    def get_eles_by_n_nodes(self, n_nodes):
        eles_by_n_nodes = {2: [], 4: [], 8: []}
        for ele in self.ele2node_tags:
            nn = len(self.ele2node_tags[ele])
            eles_by_n_nodes[nn].append(ele)
        return eles_by_n_nodes[n_nodes]

    def compute_ele_strains_and_disps(self):  # currently only available for quad elements
        import numpy as np
        rd = {}
        eles = self.get_eles_by_n_nodes(4)
        nodes = np.array([self.ele2node_tags[ele] for ele in eles]) - 1
        xd = self.x_disp[:, nodes].transpose(1, 2, 0)
        yd = self.y_disp[:, nodes].transpose(1, 2, 0)
        xc = self.coords[nodes, 0]
        yc = self.coords[nodes, 1]
        x_disp_ele, y_disp_ele = calc_quad_centroids(xc[:, :, np.newaxis] + xd, yc[:, :, np.newaxis] + yd, axis=-2)

        rd['XDISP'] = x_disp_ele - x_disp_ele[:, 0][:, np.newaxis]
        rd['YDISP'] = y_disp_ele - y_disp_ele[:, 0][:, np.newaxis]

        # return
        # nodes must be anti-clockwise
        xc0 = xc[0]
        yc0 = yc[0]
        i = 0
        for i in range(4):
            if xc0[i%4] < xc0[(i+1)%4] and xc0[(i+2)%4] > xc0[(i+3)%4] and yc0[(i+1)%4] < yc0[(i+2)%4]:  # i=bottom-left
                break
            if i == 3:
                print('WARNING could not find bottom left')
        inds = np.roll(np.arange(4), i)
        xc = xc[:, inds]
        yc = yc[:, inds]
        xd = xd[:, inds]
        yd = yd[:, inds]

        xlen = (xc[:, 1] - xc[:, 0] + xc[:, 2] - xc[:, 3]) / 2
        xdelta = (xd[:, 1] - xd[:, 0] + xd[:, 2] - xd[:, 3]) / 2
        rd['EPS_XX'] = xdelta / xlen[:, np.newaxis]
        ylen = (yc[:, 2] - yc[:, 1] + yc[:, 3] - yc[:, 0]) / 2
        ydelta = (yd[:, 2] - yd[:, 1] + yd[:, 3] - yd[:, 0]) / 2
        rd['EPS_YY'] = ydelta / ylen[:, np.newaxis]

        xd_bot = (xd[:, 1] + xd[:, 0]) / 2
        xd_top = (xd[:, 2] + xd[:, 3]) / 2
        yd_lhs = (yd[:, 3] + yd[:, 0]) / 2
        yd_rhs = (yd[:, 2] + yd[:, 1]) / 2
        rd['EPS_XY'] = ((xd_bot - xd_top) / ylen[:, np.newaxis] + (yd_rhs - yd_lhs) / xlen[:, np.newaxis]) / 2  # +ve in anti-clockwise
        rd['SSTR_MAX'] = np.sqrt(((rd['EPS_XX'] - rd['EPS_YY']) / 2) ** 2 + rd['EPS_XY'] ** 2)
        rd['EPS_VOL'] = rd['EPS_XX'] + rd['EPS_YY']
        return rd



from o3seespy.command.nd_material.base_material import NDMaterialBase


class SAniSandMS(NDMaterialBase):
    """
    The SAniSandMS NDMaterial Class

    This command is used to construct a multi-dimensional SAniSandMS material.
    """
    op_type = 'SAniSandMS'

    def __init__(self, osi, g0, nu, e_init, m_c, c_c, lambda_c, e_0, ksi, p_atm, m_yield, h_0, c_h, n_b, a_0, n_d,
                 zeta, mu0, beta, den, int_scheme=1, tan_type=0, tol_f=1.0e-7, tol_r=1.0e-7):
        r"""
        Initial method for SAniSandMS

        Note: When in elastic mode the shear modulus is equal to the shear modulus at the atmospheric pressure

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        g0: float
            Shear modulus constant
        nu: float
            Poisson ratio
        e_init: float
            Initial void ratio
        m_c: float
            Critical state stress ratio
        c_c: float
            Ratio of critical state stress ratio in extension and compression
        lambda_c: float
            Critical state line constant
        e_0: float
            Critical void ratio at p = 0
        ksi: float
            Critical state line constant
        p_atm: float
            Atmospheric pressure
        m_yield: float
            Yield surface constant (radius of yield surface in stress ratio space)
        h_0: float
            Constant parameter
        c_h: float
            Constant parameter
        n_b: float
            Bounding surface parameter, :math:`n_b \ge 0`
        a_0: float
            Dilatancy parameter
        n_d: float
            Dilatancy surface parameter :math:`n_d \ge 0`
        zeta: float
            Memory surface shrinkage parameter
        mu0: float
            Ratcheting parameter
        den: float
            Mass density of the material
        int_scheme: int, optional (default=1)
            Integration scheme type:
                * 1 = Modified Euler
                * 3 = Runge Kutta 4th order with error control (Reccommended)
        tan_type: int, optional (default=0)
            Tangent type: (appears to be unused - possibly only used in Implicit scheme)
                * 0: Elastic Tangent
                * 1: Contiuum ElastoPlastic Tangent
                * 2: Consistent ElastoPlastic Tangent
        jaco_type: int, optional (default=1)
            Jacobian type: (possibly only used in Implicit scheme)
                * 0: Finite Difference Jacobian
                * 1: Analytical Jacobian
        tol_f: float, optional (default=1.0e-7)
        tol_r: float, optional (default=1.0e-7)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.SAniSandMS(osi, g0=1110.0, nu=0.05, e_init=0.72, m_c=1.27, c_c=0.712, lambda_c=0.049, e_0=0.845,
        >>> ksi=0.27, p_atm=101.3, m_yield=0.01, h_0=5.95, c_h=1.01, n_b=2.0, a_0=1.06, n_d=1.17, zeta=0.0005, mu0=260.0, beta=1, den=1.6)
        """
        self.osi = osi
        self.g0 = float(g0)
        self.nu = float(nu)
        self.e_init = float(e_init)
        self.m_c = float(m_c)
        self.c_c = float(c_c)
        self.lambda_c = float(lambda_c)
        self.e_0 = float(e_0)
        self.ksi = float(ksi)
        self.p_atm = float(p_atm)
        self.m_yield = float(m_yield)
        self.h_0 = float(h_0)
        self.c_h = float(c_h)
        self.n_b = float(n_b)
        self.a_0 = float(a_0)
        self.n_d = float(n_d)
        self.zeta = float(zeta)  # note DM04 has z_max, c_z
        self.mu0 = float(mu0)
        self.beta = beta
        self.den = float(den)
        self.int_scheme = int(int_scheme)
        self.tan_type = int(tan_type)
        dep_jaco_type = 1
        self.tol_f = float(tol_f)
        self.tol_r = float(tol_r)
        dep_fabric_flag = 1
        dep_flow_flag = 1
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.g0, self.nu, self.e_init, self.m_c, self.c_c, self.lambda_c,
                            self.e_0, self.ksi, self.p_atm, self.m_yield, self.h_0, self.c_h, self.n_b, self.a_0,
                            self.n_d, self.zeta, self.mu0, self.beta, self.den, dep_fabric_flag, dep_flow_flag,
                            self.int_scheme, self.tan_type, dep_jaco_type,
                            self.tol_f, self.tol_r]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, stage):
        parameters = ['-material', self.tag, '-stage', stage]
        self.osi.to_process("updateMaterialStage", parameters)

    def set_material_state(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'materialState', value, ele, eles)

    def set_integration_scheme(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'IntegrationScheme', value, ele, eles)

    def set_ref_shear_modulus(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'refShearModulus', value, ele, eles)
    #
    # def set_g_mod(self, value, ele=None, eles=None):
    #     self.set_parameter(self.osi, 'ShearModulus', value, ele, eles)

    def set_nu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'poissonRatio', value, ele, eles)

    def set_void_ratio(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'voidRatio', value, ele, eles)

    def update_to_nonlinear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 1)

    def update_to_linear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 0)
